﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Xenon.KifuNarabe
{

    /// <summary>
    /// 盤上用の升目符号。「元」→「現」の形。
    /// 
    /// 「先後」、
    /// 「置き場」、
    /// 「筋、段」（明記がない場合、dstの意）、
    /// 「src筋、src段」（オプション）。
    /// 
    /// ・“同”
    /// 棋譜用の升目符号。“同”が使えます。
    /// “同”にする場合は、UNKNOWN_SUJI 筋、UNKNOWN_DAN 段に設定してください。毎回再計算します。
    /// 
    /// ・前の升：「7g7f」などにも使えるよう、src筋、src段も補助で用意。
    /// ・前の手：また、遡れるように、１つ前の升目符号へのリンクも補助で容易。
    /// 
    /// ・駒種類：「歩」「と金」など。補助で容易。
    /// </summary>
    [Serializable]
    public class Process
    {

        #region 定数
        //------------------------------------------------------------
        /// <summary>
        /// 初期設定のまま変わっていません。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int INITIALIZED_SUJI = -98;

        /// <summary>
        /// 初期設定のまま変わっていません。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int INITIALIZED_DAN = -98;

        /// <summary>
        /// 駒袋に入っていたとき。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int KOMABUKURO_SUJI = -97;

        /// <summary>
        /// 駒袋に入っていたとき。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int KOMABUKURO_DAN = -97;

        /// <summary>
        /// ボタンなので場所固定。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int BUTTON_SUJI = -96;

        /// <summary>
        /// ボタンなので場所固定。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int BUTTON_DAN = -96;

        /// <summary>
        /// 正しくない数字が入っているときです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ERROR_SUJI = -95;

        /// <summary>
        /// 正しくない数字が入っているときです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ERROR_DAN = -95;

        /// <summary>
        /// 一手前局面の駒位置を見る必要があります。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ITTEMAE_KYOKUMEN_SUJI = -94;

        /// <summary>
        /// 一手前局面の駒位置を見る必要があります。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ITTEMAE_KYOKUMEN_DAN = -94;

        /// <summary>
        /// 使わないはずの、srcのsrcです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int TUKAWANAI_SRCSRC_SUJI = -93;

        /// <summary>
        /// 使わないはずの、srcのsrcです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int TUKAWANAI_SRCSRC_DAN = -93;

        /// <summary>
        /// 符号で探し出せなかった駒の場合です。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int DA_SUJI = -92;

        /// <summary>
        /// 符号で探し出せなかった駒の場合です。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int DA_DAN = -92;
        //------------------------------------------------------------
        /// <summary>
        /// 筋は 1～9 だけ有効です。
        /// </summary>
        public const int YUKO_SUJI_MIN = 1;
        public const int YUKO_SUJI_MAX = 9;

        /// <summary>
        /// 段は 1～9 だけ有効です。
        /// </summary>
        public const int YUKO_DAN_MIN = 1;
        public const int YUKO_DAN_MAX = 9;
        //------------------------------------------------------------
        #endregion



        #region プロパティー類
        //------------------------------------------------------------
        private Sengo sengo;

        /// <summary>
        /// 先後
        /// </summary>
        public Sengo Sengo
        {
            get
            {
                return this.sengo;
            }
        }
        //------------------------------------------------------------
        private Okiba okiba;

        /// <summary>
        /// 現・駒の置き場
        /// </summary>
        public Okiba Okiba
        {
            get
            {
                return this.okiba;
            }
        }

        private Okiba srcOkiba;

        /// <summary>
        /// 元・駒の置き場
        /// </summary>
        public Okiba SrcOkiba
        {
            get
            {
                return this.srcOkiba;
            }
        }
        //------------------------------------------------------------
        private int suji;

        /// <summary>
        /// 現・筋。
        /// </summary>
        public virtual int Suji
        {
            get
            {
                return this.suji;
            }
        }

        private int srcSuji;

        /// <summary>
        /// 元・筋
        /// </summary>
        public virtual int SrcSuji
        {
            get
            {
                return this.srcSuji;
            }
        }
        //------------------------------------------------------------
        private int dan;

        /// <summary>
        /// 現・段。
        /// </summary>
        public int Dan
        {
            get
            {
                return this.dan;
            }
        }

        private int srcDan;

        /// <summary>
        /// 元・段
        /// </summary>
        public virtual int SrcDan
        {
            get
            {
                return this.srcDan;
            }
        }
        //------------------------------------------------------------
        private KomaSyurui syurui;
        /// <summary>
        /// 現・駒の種類。歩、桂…。
        /// </summary>
        public KomaSyurui Syurui
        {
            get
            {
                return this.syurui;
            }
        }

        private KomaSyurui srcSyurui;
        /// <summary>
        /// 元・駒の種類。歩、桂…。  ＜不成判定用＞
        /// </summary>
        public KomaSyurui SrcSyurui
        {
            get
            {
                return this.srcSyurui;
            }
        }
        //------------------------------------------------------------
        private KomaSyurui tottaKoma;
        /// <summary>
        /// あれば、取った駒。
        /// </summary>
        public KomaSyurui TottaKoma
        {
            get
            {
                return this.tottaKoma;
            }
        }
        //------------------------------------------------------------
        #endregion

        /// <summary>
        /// 駒用。
        /// </summary>
        /// <param name="sengo"></param>
        /// <param name="okiba"></param>
        /// <param name="srcOkiba"></param>
        /// <param name="suji"></param>
        /// <param name="dan"></param>
        /// <param name="srcSuji"></param>
        /// <param name="srcDan"></param>
        /// <param name="syurui"></param>
        /// <param name="srcSyurui"></param>
        /// <param name="previousTe"></param>
        public Process(Sengo sengo, Okiba okiba, int suji, int dan, KomaSyurui syurui)
        {
            this.sengo = sengo;
            this.okiba = okiba;
            this.srcOkiba = Okiba.NO_USE;
            this.suji = suji;
            this.dan = dan;
            this.srcSuji = Process.TUKAWANAI_SRCSRC_SUJI;
            this.srcDan = Process.TUKAWANAI_SRCSRC_DAN;
            this.syurui = syurui;
            this.srcSyurui = KomaSyurui.UNKNOWN;
            this.tottaKoma = KomaSyurui.TOTTA_KOMA_NASI;
        }

        /// <summary>
        /// 棋譜用。
        /// </summary>
        /// <param name="sengo"></param>
        /// <param name="okiba"></param>
        /// <param name="srcOkiba"></param>
        /// <param name="suji"></param>
        /// <param name="dan"></param>
        /// <param name="srcSuji"></param>
        /// <param name="srcDan"></param>
        /// <param name="syurui"></param>
        /// <param name="srcSyurui"></param>
        /// <param name="previousTe"></param>
        public Process(Sengo sengo, Okiba okiba, Okiba srcOkiba, int suji, int dan, int srcSuji, int srcDan, KomaSyurui syurui, KomaSyurui srcSyurui, KomaSyurui tottaKoma)
        {
            this.sengo = sengo;
            this.okiba = okiba;
            this.srcOkiba = srcOkiba;
            this.suji = suji;
            this.dan = dan;
            this.srcSuji = srcSuji;
            this.srcDan = srcDan;
            this.srcSyurui = srcSyurui;
            this.syurui = syurui;
            this.tottaKoma = tottaKoma;
        }


        public Process Src()
        {
            Process src = new Process(
                this.Sengo,
                this.Okiba,
                this.SrcOkiba,
                this.srcSuji,
                this.srcDan,
                Process.TUKAWANAI_SRCSRC_SUJI,
                Process.TUKAWANAI_SRCSRC_DAN,
                this.SrcSyurui,
                KomaSyurui.UNKNOWN,//KomaSyurui.FU,//TODO:不明
                KomaSyurui.UNKNOWN
                );

            return src;
        }


        #region 比較判定系
        //------------------------------------------------------------
        public bool MatchOkiba(Okiba okiba2)
        {
            return
                this.okiba == okiba2;
        }

        public bool MatchOkiba(Process m2)
        {
            return this.MatchOkiba(m2.Okiba);
        }

        public bool MatchSengo(Process m2)
        {
            return
                this.sengo == m2.sengo;
        }

        public bool MatchSuji(int suji2)
        {
            return
                // 有効な値かどうか
                Process.YUKO_SUJI_MIN<=this.suji
                && this.suji<=Process.YUKO_SUJI_MAX

                && this.suji == suji2;
        }

        public bool MatchSuji(Process m2)
        {
            return this.MatchSuji(m2.Suji);
        }

        public bool MatchDan(int dan2)
        {
            return
                // 有効な値かどうか
                Process.YUKO_DAN_MIN <= this.dan
                && this.dan <= Process.YUKO_DAN_MAX

                && this.dan == dan2;
        }

        public bool MatchDan(Process m2)
        {
            return this.MatchDan(m2.Dan);
        }
        //------------------------------------------------------------
        public bool MatchMasu(Object obj2)
        {
            bool result = false;

            if(obj2 is Process)
            {
                Process m2 = (Process)obj2;
                result = this.MatchSuji(m2) && this.MatchDan(m2);
            }

            return result;
        }
        //------------------------------------------------------------
        public bool MatchKoma(Object obj2)
        {
            bool result = false;

            if (obj2 is Process)
            {
                Process m2 = (Process)obj2;
                result = this.MatchSengo(m2)
                    && this.MatchSuji(m2)
                    && this.MatchDan(m2);
            }

            return result;
        }
        //------------------------------------------------------------
        /// <summary>
        /// 一致判定：   升目符号同士
        /// </summary>
        /// <param name="masu2"></param>
        /// <returns></returns>
        public bool Match(Process masu2)
        {
            return
                this.MatchOkiba(masu2)
                && this.MatchSuji(masu2)
                && this.MatchDan(masu2);
        }
        //------------------------------------------------------------
        public bool ExistsIn(List<Process> masu2Arr)
        {
            bool matched = false;

            foreach (Process m in masu2Arr)
            {
                if (
                    null != m
                    && this.MatchSengo(m) //2014-07-21 先後も見るように追記。
                    && this.MatchSuji(m)
                    && this.MatchDan(m)
                    )
                {
                    matched = true;
                    break;
                }
            }

            return matched;
        }
        //------------------------------------------------------------
        #endregion



        /// <summary>
        /// １マス上、のように指定して、マスを取得します。
        /// </summary>
        /// <param name="offsetSuji"></param>
        /// <param name="offsetDan"></param>
        /// <returns></returns>
        public Process Offset(Sengo sengo, int offsetSuji, int offsetDan)
        {
            return new Process(
                sengo, 
                this.Okiba,
                this.SrcOkiba,
                this.Suji + offsetSuji,
                this.Dan + offsetDan,
                this.SrcSuji,
                this.SrcDan,
                this.Syurui,
                this.SrcSyurui,
                this.TottaKoma
                );
        }


        public string ToText()
        {
            StringBuilder sb = new StringBuilder();


            if (this.IsDaAction)
            {
                sb.Append(Converter.SyuruiToSfenDa(this.SrcSyurui));
                sb.Append("*");
            }
            else
            {
                //------------------------------------------------------------
                // 元筋
                //------------------------------------------------------------
                string strSrcSuji;
                switch (this.SrcSuji)
                {
                    case BUTTON_SUJI:
                        strSrcSuji = "ボ";
                        break;
                    case ERROR_SUJI:
                        strSrcSuji = "エ";
                        break;
                    case ITTEMAE_KYOKUMEN_SUJI:
                        strSrcSuji = "前";
                        break;
                    case KOMABUKURO_SUJI:
                        strSrcSuji = "袋";
                        break;
                    case DA_SUJI:
                        strSrcSuji = "＜srcOkiba="+this.SrcOkiba+"＞";
                        //strSrcSuji = Converter.SyuruiToSfenDa(this.Syurui);//打のとき。
                        break;
                    case TUKAWANAI_SRCSRC_SUJI:
                        strSrcSuji = "ソ元筋";
                        break;
                    case INITIALIZED_SUJI:
                        strSrcSuji = "Ｉ";
                        break;
                    default:
                        strSrcSuji = this.SrcSuji.ToString();
                        break;
                }
                sb.Append(strSrcSuji);

                //------------------------------------------------------------
                // 元段
                //------------------------------------------------------------
                string strSrcDan;
                switch (this.SrcDan)
                {
                    case BUTTON_DAN:
                        strSrcDan = "ボ";
                        break;
                    case ERROR_DAN:
                        strSrcDan = "エ";
                        break;
                    case ITTEMAE_KYOKUMEN_DAN:
                        strSrcDan = "前";
                        break;
                    case KOMABUKURO_DAN:
                        strSrcDan = "袋";
                        break;
                    //case DA_DAN:
                    //    strSrcDan = "*";//仕様：打のとき、元段は *表記 になります。
                    //    break;
                    case TUKAWANAI_SRCSRC_DAN:
                        strSrcDan = "ソ元段";
                        break;
                    case INITIALIZED_DAN:
                        strSrcDan = "Ｉ段";
                        break;
                    default:
                        strSrcDan = Converter.IntToAlphabet(this.SrcDan);
                        break;
                }
                sb.Append(strSrcDan);
            }

            //------------------------------------------------------------
            // 現筋
            //------------------------------------------------------------
            string strSuji;
            switch (this.Suji)
            {
                case BUTTON_SUJI:
                    strSuji = "ボ";
                    break;
                case ERROR_SUJI:
                    strSuji = "エ";
                    break;
                case ITTEMAE_KYOKUMEN_SUJI:
                    strSuji = "前";
                    break;
                case KOMABUKURO_SUJI:
                    strSuji = "袋";
                    break;
                case DA_SUJI:
                    strSuji = "Ｄ";//エラー
                    break;
                case TUKAWANAI_SRCSRC_SUJI:
                    strSuji = "ソ現筋";
                    break;
                case INITIALIZED_SUJI:
                    strSuji = "Ｉ";
                    break;
                default:
                    strSuji = this.Suji.ToString();
                    break;
            }
            sb.Append(strSuji);


            //------------------------------------------------------------
            // 現段
            //------------------------------------------------------------
            string strDan;
            switch (this.Dan)
            {
                case BUTTON_DAN:
                    strDan = "ボ";
                    break;
                case ERROR_DAN:
                    strDan = "エ";
                    break;
                case ITTEMAE_KYOKUMEN_DAN:
                    strDan = "前";
                    break;
                case KOMABUKURO_DAN:
                    strDan = "袋";
                    break;
                case DA_DAN:
                    strDan = "Ｄ";//エラー
                    break;
                case TUKAWANAI_SRCSRC_DAN:
                    strDan = "ソ現段";
                    break;
                case INITIALIZED_DAN:
                    strDan = "Ｉ段";
                    break;
                default:
                    strDan = Converter.IntToAlphabet(this.Dan);
                    break;
            }
            sb.Append(strDan);


            //------------------------------------------------------------
            // 成
            //------------------------------------------------------------
            if (this.IsNatta_Process)
            {
                sb.Append("+");
            }

            return sb.ToString();
        }

        public override string ToString()
        {
            return this.ToText();
        }


        /// <summary>
        /// 移動前と、移動後の場所が異なっていれば真。
        /// </summary>
        /// <returns></returns>
        public bool DoneMove()
        {
            return !this.Match(this.Src());
        }

        /// <summary>
        /// 不成ケース
        /// </summary>
        /// <returns></returns>
        public KomaSyurui ToFunariCase()
        {
            return Converter.SyuruiToFunariCase(this.Syurui);
        }

        /// <summary>
        /// 成ケース
        /// </summary>
        /// <returns></returns>
        public KomaSyurui ToNariCase()
        {
            return Converter.SyuruiToNariCase(this.Syurui);
        }

        /// <summary>
        /// 駒の表面の文字。
        /// </summary>
        public string Text_Label
        {
            get
            {
                return Converter.SyuruiToLabel(this.Syurui);
            }
        }

        /// <summary>
        /// 成り
        /// </summary>
        public bool Nari
        {
            get
            {
                return Converter.SyuruiToNari(this.Syurui);
            }
        }

        /// <summary>
        /// 成った
        /// </summary>
        /// <returns></returns>
        public bool IsNatta_Process
        {
            get
            {
                return !Converter.IsSyuruiNari(this.SrcSyurui) && Converter.IsSyuruiNari(this.Syurui);
            }
        }

        /// <summary>
        /// “打” ＜アクション時＞
        /// </summary>
        /// <returns></returns>
        public bool IsDaAction
        {
            get
            {
                return
                    Okiba.SHOGIBAN != this.SrcOkiba
                    && Okiba.NO_USE != this.SrcOkiba;//初期配置から移動しても、打にはしません。
            }
        }

        /// <summary>
        /// 一致判定：　先後、駒種類
        /// </summary>
        /// <returns></returns>
        public bool OnajiKoma(Button_Koma koma2)
        {
            return koma2 != null
                && this.Sengo == koma2.KomaProperty.Sengo
                && this.Syurui == koma2.KomaProperty.Syurui;
        }

        /// <summary>
        /// 不一致判定：　先後、駒種類  が、自分と同じものが　＜ひとつもない＞
        /// </summary>
        /// <returns></returns>
        public bool NeverOnaji(params List<Button_Koma>[] kGroupArgs)
        {
            bool unmatched = true;

            foreach (List<Button_Koma> kGroup in kGroupArgs)
            {
                foreach (Button_Koma k in kGroup)
                {

                    if (
                        null != k
                        && this.Sengo == k.KomaProperty.Sengo
                        && this.Syurui == k.KomaProperty.Syurui
                        )
                    {
                        // １つでも一致するものがあれば、終了します。
                        unmatched = false;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return unmatched;
        }

        /// <summary>
        /// 一致判定：　先後、駒種類  が、自分と同じものが　＜ひとつ以上ある＞
        /// </summary>
        /// <returns></returns>
        public bool ExistsOnaji(params Button_Koma[][] kGroupArgs)
        {
            bool matched = false;

            foreach (Button_Koma[] kGroup in kGroupArgs)
            {
                foreach (Button_Koma btnKoma1 in kGroup)
                {
                    if (null != btnKoma1 && this.Sengo == btnKoma1.KomaProperty.Sengo && this.Syurui == btnKoma1.KomaProperty.Syurui)
                    {
                        // １つでも一致するものがあれば、終了します。
                        matched = true;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return matched;
        }

        /// <summary>
        /// 一致判定：　先後、駒種類  が、自分と同じもの。なければヌル。
        /// </summary>
        /// <returns></returns>
        public Button_Koma MatchFirst(params Button_Koma[][] kGroupArgs)
        {
            Button_Koma matched = null;

            foreach (Button_Koma[] kGroup in kGroupArgs)
            {
                foreach (Button_Koma k1 in kGroup)
                {
                    if (null != k1 && this.Sengo == k1.KomaProperty.Sengo && this.Syurui == k1.KomaProperty.Syurui)
                    {
                        // １つでも一致するものがあれば、終了します。
                        matched = k1;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return matched;
        }

        /// <summary>
        /// 一致判定：   場所指定
        /// </summary>
        /// <returns></returns>
        public bool Contains(Process masu)
        {
            bool matched = false;

            if (masu.Match(this))
            {
                matched = true;
            }

            return matched;
        }

        /// <summary>
        /// 相手陣に入っていれば真。
        /// </summary>
        /// <returns></returns>
        public bool InAitejin()
        {
            return
                // 後手は 7,8,9 段。
                (Sengo.GOTE == this.Sengo && 7 <= this.Dan)
                // 先手は 1,2,3 段。
                || (Sengo.SENTE == this.Sengo && this.Dan <= 3);
        }

    }

}
